// CUDA runtime
#include <cuda_runtime.h>

#include <cuda_profiler_api.h>

// Helper functions and utilities to work with CUDA
#include <helper_functions.h>

/**********************************************
 * Check whether we read back the same input
 * The double check is just for debug purposes.
 * We can comment it out when benchmarking the time.
 **********************************************/
#define GPU_DEBUG

/******************************************
 * Device function declaration
 *****************************************/
__global__
void layer1_init_bias(float* d_y, float* d_bias, int r);
__global__
void layer1_feature_maps(float* d_y, unsigned char* d_in_layer, float* d_weight, int r);
__global__
void layer1_sigmoid(float* d_y, unsigned char* d_out_layer, int r);

void cuda_profiler_init(){
  cudaError_t error;
  /************************************
   * Profiler has two output formats.
   * cudaCSV : CSV format
   * cudaKeyValuePair : readable text
   ***********************************/
  error = cudaProfilerInitialize("profiler_config.txt", "profiler_output%d.csv", cudaCSV);
  if (error != cudaSuccess){
    printf("cudaProfilerInitialize returned error code %d, line(%d)\n", error, __LINE__);
    printf("CUDA error: %s\n", cudaGetErrorString(error));
    exit(EXIT_FAILURE);
  }else{
    printf("cudaProfilerInitialize success.\n");
  }
}

void cuda_profiler_start(){
  cudaError_t error;
  error = cudaProfilerStart();
  if (error != cudaSuccess){
    printf("cudaProfilerStart returned error code %d, line(%d)\n", error, __LINE__);
    printf("CUDA error: %s\n", cudaGetErrorString(error));
    exit(EXIT_FAILURE);
  }else{
    printf("cudaProfilerStart success.\n");
  }
}

void cuda_profiler_stop(){
  cudaError_t error;
  error = cudaProfilerStop();
  if (error != cudaSuccess){
    printf("cudaProfilerStop returned error code %d, line(%d)\n", error, __LINE__);
    printf("CUDA error: %s\n", cudaGetErrorString(error));
    exit(EXIT_FAILURE);
  }else{
    printf("cudaProfilerStop success.\n");
  }
}

/************************************************************************************
 * Input   : input image, pointer to output result, coefficients bias and weights
 * Output  : neuron outputs of the feature maps represented as an image
 * Procedure: perform feed forward computation through the feature extraction layers
     *******************************************************************************/
void cuda_convolution_layer1(unsigned char in_layer[], unsigned char out_layer[],
			     const float bias[], const float weight[]) {

  unsigned int i;

  /*********************************
   * allocate device memory on GPU
   *********************************/

  unsigned int size_y = 6*358*638;
  unsigned int mem_size_y = sizeof(float) * size_y;
  float *d_y;

  unsigned int size_bias = 6;
  unsigned int mem_size_bias = sizeof(float) * size_bias;
  float *d_bias;

  unsigned int size_weight = 6*36;
  unsigned int mem_size_weight = sizeof(float) * size_weight;
  float *d_weight;

  unsigned int size_in_layer = 720*1280;
  unsigned int mem_size_in_layer = sizeof(unsigned char) * size_in_layer;
  unsigned char *d_in_layer;

  unsigned int size_out_layer = 6*358*638;
  unsigned int mem_size_out_layer = sizeof(unsigned char) * size_out_layer;
  unsigned char *d_out_layer;

  cudaError_t error;

  // Setup execution parameters 
  dim3 threads;
  dim3 grid;

  /* for debug purposes, check the read back */
#ifdef GPU_DEBUG
  int error_cnt;
  unsigned char cuda_in_layer[720*1280];
  float cuda_weight[6*36];
  float cuda_bias[6];
#endif

  /********************************
   * Allocate device memory on GPU.
   * Check the first cudaMalloc error,
   * in case GPU is busy.
   ********************************/
  error = cudaMalloc((void **) &d_y, mem_size_y);
  /* Check the error code of the first CUDA API call */
  if (error != cudaSuccess){
    printf("cudaMalloc returned error code %d, line(%d)\n", error, __LINE__);
    printf("CUDA error: %s\n", cudaGetErrorString(error));
    exit(EXIT_FAILURE);
  }else{
    printf("cudaMalloc success.\n");
  }
  /* if no error for the first cudaMalloc, continue other cudaMalloc */
  error = cudaMalloc((void **) &d_in_layer, mem_size_in_layer);
  error = cudaMalloc((void **) &d_bias, mem_size_bias);
  error = cudaMalloc((void **) &d_weight, mem_size_weight);
  error = cudaMalloc((void **) &d_out_layer, mem_size_out_layer);

  /*********************************************
   * copy data from host (CPU) to device (GPU)
   ********************************************/
  error = cudaMemcpy(d_in_layer, in_layer, mem_size_in_layer, cudaMemcpyHostToDevice);
  error = cudaMemcpy(d_bias, bias, mem_size_bias, cudaMemcpyHostToDevice);
  error = cudaMemcpy(d_weight, weight, mem_size_weight, cudaMemcpyHostToDevice);

  /* Synchronize all the cudaMemcpy API before doing the actual computation */
  cudaDeviceSynchronize();

  /*********************************************
   * Layer 1, Step 1: 
   * init values of feature maps at bias value 
   ********************************************/
  /* 16*16 threads per block */
  /* NOTE: threads per block limit is 1024 for GTX 460 */
  /* NOTE: if you use another GPU, check the deviceQuery */
  threads.x = 16;
  threads.y = 16;
  threads.z = 1;
  /***********************************************
   * The layer size is not diviadable by 16.
   * Anyway, allocate threads in multiple of 16,
   * but mask out threads in the kernel.
   **********************************************/
  grid.x = (638/16)+1;
  grid.y = (358/16)+1;
  grid.z = 1;
  /* Loop over the feature maps with 6 seperate launches */
  /* It is to prevent time out of a large kernel. */
  for(i = 0; i < 6; i++){
    layer1_init_bias<<<grid,threads>>>(d_y, d_bias, i);
  }

  /* Just in case, put a sync here */
  cudaDeviceSynchronize();

  /*********************************************
   * Layer 1, Step 2: 
   * loop over output feature maps
   ********************************************/
  /* 8*8 threads per block */
  threads.x = 8;
  threads.y = 8;
  threads.z = 1;
  /***********************************************
   * The layer size is not diviadable by 8 either.
   * Mask out extra threads in the kernel.
   **********************************************/  
  grid.x = (638/8)+1;
  grid.y = (358/8)+1;
  grid.z = 1;
  /* Loop over each feature map */
  for(i = 0; i < 6; i++){
    layer1_feature_maps<<<grid,threads>>>(d_y, d_in_layer, d_weight, i);
  }

  /* Just in case, put a sync here */
  cudaDeviceSynchronize();

  /*********************************************
   * Layer 1, Step 3: 
   * sigmoid activation function
   ********************************************/
  threads.x = 16;
  threads.y = 16;
  threads.z = 1;
  grid.x = (638/16)+1;
  grid.y = (358/16)+1;
  grid.z = 1;
  for(i = 0; i < 6; i++){
    layer1_sigmoid<<<grid,threads>>>(d_y, d_out_layer, i);
  }

  /* Just in case, put a sync here */
  cudaDeviceSynchronize();

  /* Read back the output from device (GPU) to host (CPU) */
  error = cudaMemcpy(out_layer, d_out_layer, mem_size_out_layer, cudaMemcpyDeviceToHost);
  /* For debug purposes, also check whether the input is indeed correct */
#ifdef GPU_DEBUG
  error = cudaMemcpy(cuda_bias, d_bias, mem_size_bias, cudaMemcpyDeviceToHost);
  error = cudaMemcpy(cuda_weight, d_weight, mem_size_weight, cudaMemcpyDeviceToHost);
  error = cudaMemcpy(cuda_in_layer, d_in_layer, mem_size_in_layer, cudaMemcpyDeviceToHost);
#endif

  /* Just in case, put a sync here */
  cudaDeviceSynchronize();

  /* release device memory */
  cudaFree(d_y);
  cudaFree(d_in_layer);
  cudaFree(d_bias);
  cudaFree(d_weight);
  cudaFree(d_out_layer);

  /**********************************************
   * Check whether we read back the same input
   * The double check is just for debug purposes.
   * We can comment it out when benchmarking the time.
   **********************************************/
#ifdef GPU_DEBUG
  /* check bias read back */
  error_cnt = 0;
  for(i = 0; i < size_bias; i++){
    if(cuda_bias[i]!=bias[i]){
      error_cnt++;
    }
  }
  if(error_cnt == 0){
    printf("Bias read back: passed.\n");
  }else{
    printf("Bias read back has eror. Number of error: %d\n", error_cnt);
  }
  /* check weight read back */
  error_cnt = 0;
  for(i = 0; i < size_weight; i++){
    if(cuda_weight[i]!=weight[i]){
      error_cnt++;
    }
  }
  if(error_cnt == 0){
    printf("Weight read back: passed.\n");
  }else{
    printf("Weight read back has error. Number of error: %d\n", error_cnt);
  }
  /* check in_layer read back */
  error_cnt = 0;
  for(i = 0; i < 720*128; i++){
    if( cuda_in_layer[i] != in_layer[i] ){
      error_cnt++;
    }
  }
  if(error_cnt == 0){
    printf("In_layer read back: passed.\n");
  }else{
    printf("In_layer read back has error. Number of error: %d\n", error_cnt);
  }
#endif

}


/*********************************************
 * GPU kernel
 * Layer 1, Step 1: 
 * init values of feature maps at bias value 
 ********************************************/
__global__
void layer1_init_bias(float* d_y, float* d_bias, int r){
  // Thread index 
  int tx = threadIdx.x;
  int ty = threadIdx.y;
  // Block index
  int bx = blockIdx.x;
  int by = blockIdx.y;
  /* row and column index */
  int m, n;
  /* get the index for each thread */
  m = by*16+ty;
  n = bx*16+tx;
  /* mask out threads that are out of bound */
  if(m<358 && n<638){
    d_y[r*358*638+m*638+n]=d_bias[r];
  }

}

/*********************************************
 * GPU kernel
 * Layer 1, Step 2: 
 * loop over output feature maps
 ********************************************/
__global__
void layer1_feature_maps(float* d_y, unsigned char* d_in_layer, float* d_weight, int r){
  // Thread index 
  int tx = threadIdx.x;
  int ty = threadIdx.y;
  // Block index
  int bx = blockIdx.x;
  int by = blockIdx.y;
  /* row and column index */
  int m,n;
  /* base address of the d_in_layer */
  int d_in_layer_base;
  /* convolution index */
  int l, k;
  /* local register for convolution */
  float accu;
  /* pack 4 unsigned char into each load */
  uchar4 uchar4_tmp;

  /********************************************
   * Shared memory for the d_weight.
   * The weight is float, with stride of 1.
   * Therefore, the footprint is simply 6*6.
   ********************************************/
  __shared__ float s_weight[6*6];

  /********************************************
   * Shared memory for a tile of d_in_layer.
   * There are 8*8 thread, each has a stride of 2.
   * The convolution has a window size of 6,
   * from which 5 elements need to be padded.
   * Thus, the memory footprint is (8*2+5)*(8*2+5).
   * To easily unpack one uchar4 to four unsigned int,
   * the column is padded to multiple of 4.
   * Therefore, there are (8*2+8) columns, (8*2+5) rows.
   *
   * In short, the tile has 21 rows and 24 columns.
   ********************************************/
  __shared__ unsigned int s_in_layer[(8*2+8)*(8*2+5)];

  /* row and column index of each thread */
  m = by*8+ty;
  n = bx*8+tx;
  /******************************************************
   * Base address of the tile in d_in_layer.
   * This tile has locality for threads of the same block.
   * Therefore, this tile will be loaded to shared mem.
   *****************************************************/
  d_in_layer_base = by*8*2*1280+bx*8*2;

  /*********************************************
   * Load a tile of d_weight to shared memory.
   * Each thread loads one float.
   *********************************************/
  /* Mask out some of the 8*8 threads to fit tile size */
  /* The tile has 6 columns and 6 rows */
  if(tx<6 && ty<6){
    s_weight[ty*6+tx] = d_weight[r*36+ty*6+tx];
  }

  /*********************************************
   * Load a tile of d_in_layer to shared memory.
   * Each thread loads four unsigned char,
   * and unpack they to four unsigned int.
   *********************************************/
  /* Mask out some of the 8*8 threads to fit tile size */
  /* The tile has 24 columns. */
  /* Thus, let 6 threads each load 4 element. */
  if(tx<6){
    /* the first sub-tile of 8 rows and (6*4) columns */
    uchar4_tmp = ((uchar4*)d_in_layer)[ (d_in_layer_base + ty*1280 + tx*4) / 4];
    s_in_layer[ty*24+tx*4] = uchar4_tmp.x;
    s_in_layer[ty*24+tx*4+1] = uchar4_tmp.y;
    s_in_layer[ty*24+tx*4+2] = uchar4_tmp.z;
    s_in_layer[ty*24+tx*4+3] = uchar4_tmp.w;
    /* the next sub-tile of 8 rows and (6*4) columns */
    uchar4_tmp = ((uchar4*)d_in_layer)[ (d_in_layer_base + (8+ty)*1280 + tx*4) / 4];
    s_in_layer[(8+ty)*24+tx*4] = uchar4_tmp.x;
    s_in_layer[(8+ty)*24+tx*4+1] = uchar4_tmp.y;
    s_in_layer[(8+ty)*24+tx*4+2] = uchar4_tmp.z;
    s_in_layer[(8+ty)*24+tx*4+3] = uchar4_tmp.w;
    /* the last sub-tile of 5 rows and (6*4) columns */
    if(ty<5){
      uchar4_tmp = ((uchar4*)d_in_layer)[ (d_in_layer_base + (16+ty)*1280 + tx*4) / 4];
      s_in_layer[(16+ty)*24+tx*4] = uchar4_tmp.x;
      s_in_layer[(16+ty)*24+tx*4+1] = uchar4_tmp.y;
      s_in_layer[(16+ty)*24+tx*4+2] = uchar4_tmp.z;
      s_in_layer[(16+ty)*24+tx*4+3] = uchar4_tmp.w;
    }
  }
  /* Finished loading whole tile of 21 rows and 24 columns. */

  /*********************************************************
   * IMPORTANT: sync!
   * Otherwise some threads will convolute with faulty data.
   ********************************************************/
  __syncthreads();

  /* mask out threads that are out of bound */
  if(m<358 && n<638){
    /* load the bias from global mem to register */
    accu = d_y[r*358*638+m*638+n];
    /* the actual convolution */
    for(k=0; k<6; k++){
      for(l=0; l<6; l++){
	/* Use the shared memory s_in_layer and s_weight */
	accu += s_in_layer[(ty*2+k)*24+(tx*2+l)] * s_weight[k*6+l];
      }
    }
    /* write back from register to global mem */
    d_y[r*358*638+m*638+n] = accu;
  }

}

/*********************************************
 * GPU kernel
 * Layer 1, Step 3: 
 * sigmoid activation function
 ********************************************/
__global__
void layer1_sigmoid(float* d_y, unsigned char* d_out_layer, int r){
  // Thread index 
  int tx = threadIdx.x;
  int ty = threadIdx.y;
  // Block index
  int bx = blockIdx.x;
  int by = blockIdx.y;
  /* row and column index of each thread */
  int m, n;
  /* get the index for each thread */
  m = by*16+ty;
  n = bx*16+tx;
  /* mask out threads that are out of bound */
  if(m<358 && n<638){
    d_out_layer[r*358*638+m*638+n]=(unsigned char)(255.999f/(1.0+expf(-1.0*d_y[r*358*638+m*638+n]/256.0)));
  }

}
